/* Copyright (C) 2004 GFRN systems

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The latest version of this program may be found at
   http://CQiNet.sourceforge.net

   $Log: avrs.c,v $
   Revision 1.5  2008/02/26 17:34:44  wb6ymh
   Added dmalloc support.

   Revision 1.4  2007/11/13 15:07:38  wb6ymh
   1. Restored logging of AVRS string, but only in debug mode.
   2. Added code to GenAVRS to round negative values for antenna height and
      gain to zero.

   Revision 1.3  2007/04/27 14:27:30  wb6ymh
   Removed left over debug logging from GenAVRS.

   Revision 1.2  2005/01/08 23:55:52  wb6ymh
   Added rounding to frequency conversion.

   Revision 1.1  2004/11/29 00:29:14  wb6ymh
   Initial import - report repeater/link location and parameters to the
   EchoLink servers.

   
EchoLink sends a packet once every 10 minutes to aprs.echolink.org.  It also
sends a packet when the node status changes (i.e., when somebody connects or
disconnects, or the node comes online or goes offline).

The packet is an RTP control packet, and thus is sent to port 5199.

The CNAME portion is the node callsign (including -L/-R), followed by a
slash, followed by a "password", which is actually a hash of the callsign.
However, at this time, the password field is ignored by the server; a future
release of the server will verify the authenticity of the packet by its
source IP address.  So the callsign can be repeated in this second field.

The format of the LOC portion is as follows:


 )EL-nnnnnn!DDMM.HHNXDDDMM.HHW0PHGphgd/FFFFFF/PPPyyyyyyyyyyzzzzzzzz

 Where:

 nnnnnn           is the 6 character node Name or Callsign (right-padded
                  with spaces to 6 chars)
 DDMM.HHN DDDMM.HHW  is the LAT/LONG of the node, using degrees, minutes,
                     and 100ths of minutes, and E/W/N/S
X                    is the letter "E", for EchoLink
 0                   is a "zero",the ICON character
p                    is the power (encoded)
h                    is the antenna height (encoded)
g                    is the antenna gain (encoded)
d                    is the antenna directivity (encoded)
 FFFFFF              is the frequency in KHz
 /PPP                is the PL in whole number Hz
 yyyyyyyyyy          Second line on a D7 (11th to 20th on a D700)
 zzzzzzzz            last 8 not shown on a D7 but are on the D700

 Formats for the variable field YYYYYYYYYY and ZZZZZZZZ:

    Offline:    " Off @HHMM........"  can contain some comment
    On-Line:    " On @ HHMM........"  Time it returned to ON-Line
    Busy:       " Busy HHMM........"  Time it went busy
    Connected:  "=NNnnnnnnn at HHMM"  Shows call/Name of connection

The encoding of the p/h/g/d fields is as follows:

     DIGITS   0  1  2   3   4   5   6    7    8    9
     -------------------------------------------------------------------
     POWER    0, 1, 4,  9, 16, 25, 36,  49,  64,  81  watts  SQR(P)
     HEIGHT  10,20,40, 80,160,320,640,1280,2560,5120  feet   LOG2(H/10)
     GAIN     0, 1, 2,  3,  4,  5,  6,   7,   8,   9  dB
     DIR   omni,45,90,135,180,225,270, 315, 360,   .  deg    (D/45)

For example:   // )EL-K1RFD !4118.52NE07329.97W0PHG7430/144300/000 On @0444
Rdgfld
*/

#include "common.h"
#ifndef _WIN32
   // FreeBSD, Linux, etc..
   #include <stdio.h>
   #include <stdlib.h>
   #ifdef TIME_WITH_SYS_TIME
      #include <sys/time.h>
      #include <time.h>
   #else
      #ifdef HAVE_SYS_TIME_H
         #include <sys/time.h>
      #else
         #include <time.h>
      #endif
   #endif
   #include <unistd.h>
   #include <sys/socket.h>
   #include <netinet/in.h>
   #include <string.h>
   #include <math.h>
   #include <ctype.h>
#else
   // Windoze
   #include <stdio.h>
   #include <time.h>
   #include <winsock2.h>
   #include <math.h>
#endif

#include "rtp.h"
#include "main.h"
#include "configvars.h"
#include "ilink.h"
#include "dirclient.h"
#include "conference.h"
#include "avrs.h"

#ifdef USE_DMALLOC
#include "dmalloc.h"
#endif

void GenAVRS(AVRS_State State,char *Connected2,char *Buf,int BufLen)
{
   static AVRS_State LastState = -1;
   static time_t StateChangeTime;
   int Wrote = 0;
   int Power;
   int Height;
   int Gain;
   int Dir;
   int Lat;
   int LatMin;
   int LatHunders;
   int Long;
   int LongHunders;
   int LongMin;
   int Freq;
   int Pl;
   char  NorthSouth;
   char  EastWest;
   char *cp;
   struct tm *pTm;
   char  SepChar;
   char  *Callsign;
   float x;

   if(State != LastState) {
   // New state, save the time
      LastState = State;
      time(&StateChangeTime);
   }

   pTm = localtime(&StateChangeTime);

   if((cp = strchr(ConferenceCall,'-')) != NULL) {
   // Remove -L or -R from conference callsign for now
      *cp = 0;
   }

   Callsign = strdup(ConferenceCall);

   if(cp != NULL) {
   // Restore -L or -R to conference callsign
      *cp = '-';
   }

   cp = Callsign;
   while(*cp) {
   // Convert to lower case
      *cp = tolower(*cp);
      cp++;
   }

   if(Latitude > 0.0) {
   // North
      NorthSouth = 'N';
      x = Latitude;
   }
   else {
   // South
      NorthSouth = 'S';
      x = -Latitude;
   }

   Lat = (int) x;
   x -= Lat;
   x *= 60.0;
   LatMin = (int) x;
   x -= LatMin;
   x *= 100.0;
   LatHunders = (int) (x + 0.5);

   if(Longitude > 0.0) {
   // East
      EastWest = 'E';
      x = Longitude;
   }
   else {
   // West
      x = -Longitude;
      EastWest = 'W';
   }

   Long = (int) x;
   x -= Long;
   x *= 60.0;
   LongMin = (int) x;
   x -= LongMin;
   x *= 100.0;
   LongHunders = (int) (x + 0.5);

   Power = (int) (sqrt(TxPower) + 0.5);
   if(Power > 9) {
      Power = 9;
   }

   Height = (int) (log(Haat/10.0) / log(2.0) + 0.5);
   if(Height > 9) {
      Height = 9;
   }

   if(Height < 0) {
      Height = 0;
   }

   Gain = (int) (AntGain + 0.5);
   if(Gain > 9) {
      Gain = 9;
   }

   if(Gain < 0) {
      Gain = 0;
   }

   Dir = (int) ((AntDir / 45.0) + 0.5);
   Freq = (int) ((Frequency * 1e3) + 0.5);
   Pl = (int) CTCSS;

   Wrote = snprintf(Buf,BufLen,
                    ")EL-%6s!%02d%02d.%02d%cE%03d%02d.%02d%c0",Callsign,
                    Lat,LatMin,LatHunders,NorthSouth,
                    Long,LongMin,LongHunders,EastWest);

   Wrote += snprintf(&Buf[Wrote],BufLen - Wrote,"PHG%d%d%d%d/%06d/%03d",
                     Power,Height,Gain,Dir,Freq,Pl);
   
   switch(State) {
      case AVRS_OFF:
         Wrote += snprintf(&Buf[Wrote],BufLen - Wrote," Off @");
         break;

      case AVRS_ON:
         Wrote += snprintf(&Buf[Wrote],BufLen - Wrote," On  @");
         break;

      case AVRS_BUSY:
         Wrote += snprintf(&Buf[Wrote],BufLen - Wrote," Busy ");
         break;

      case AVRS_CONNECTED:
         if((bEchoIrlpMode && ConferenceClients > 2) || 
            (!bEchoIrlpMode && ConferenceClients > 1))
         {  // multiple connections
            SepChar = '+';
         }
         else {
         // Single connection
            SepChar = '=';
         }

         Wrote += snprintf(&Buf[Wrote],BufLen - Wrote,"%c%s at ",SepChar,
                           Connected2);
         break;
   }

   Wrote += snprintf(&Buf[Wrote],BufLen - Wrote,"%02d%02d\r",
                     pTm->tm_hour,pTm->tm_min);

   DPRINTF(("GenAVRS(): sending string:\n%s\n",Buf));

   free(Callsign);
}


int CreateAvrsPacket(AVRS_State State,char *Connected2,char *Buf,int BufLen)
{
   char PadCount;
   char *pFirstData;
   union {
      char *cp;
      rtcp_t *p;
   } u;

   union {
      char *cp;
      rtcp_sdes_item_t *p;
   } Item;
   int CallLen;

   u.cp = Buf;

   u.p->common.version = ILINK_RTP_VERSION;
   u.p->common.p = 0;
   u.p->common.count = 0;
   u.p->common.pt = RTCP_RR;
   u.p->common.length = htons(1);
   u.p->r.rr.ssrc = 0;

   u.cp += sizeof(u.p->common) + sizeof(u.p->r.rr.ssrc);

   u.p->common.version = ILINK_RTP_VERSION;
   u.p->common.p = 1;
   u.p->common.count = 1;
   u.p->common.pt = RTCP_SDES;

   u.p->r.sdes.src = 0;
   pFirstData = (char *) &u.p->r.sdes.src;

   Item.p = &u.p->r.sdes.item[0];
   Item.p->type = RTCP_SDES_CNAME;
   CallLen = strlen(ConferenceCall);
   Item.p->length = (u_int8) (1 + CallLen + CallLen);
   memcpy(Item.p->data,ConferenceCall,CallLen);
   Item.p->data[CallLen] = '/';
// Eventually hash of the Echolink password
   memcpy(&Item.p->data[CallLen+1],ConferenceCall,CallLen);

   Item.cp = Item.cp + sizeof(rtcp_sdes_item_t) + Item.p->length - 1;
   Item.p->type = RTCP_SDES_LOC;

   GenAVRS(State,Connected2,Item.p->data,BufLen - (Item.cp - Buf));
   Item.p->length = (u_int8) strlen(Item.p->data);

   Item.cp = Item.cp + sizeof(rtcp_sdes_item_t) + Item.p->length - 1;
   Item.p->type = RTCP_SDES_END ;
   Item.p->length = 0;
   Item.cp = Item.cp + sizeof(rtcp_sdes_item_t) + Item.p->length - 1;
   PadCount = 4 - ((Item.cp - pFirstData) % 4);

   if(PadCount > 0 && PadCount < 4) {
   // Padding is needed
      u.p->common.p = 1;
      if(PadCount > 1) {
         memset(Item.cp,0,PadCount-1);
      }
      Item.cp[PadCount-1] = PadCount;
      Item.cp += PadCount;
   }
   else {
   // No padding needed
      u.p->common.p = 0;
   }
   u.p->common.length = htons((u_short) ((Item.cp - pFirstData)/4));
   return Item.cp - Buf;
}

